#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>

// Old Turbo/Borland C/C++ doesn't have the S_I?USR defines:
#if !defined(S_IRUSR)
  #define S_IRUSR S_IREAD
  #define S_IWUSR S_IWRITE
  #define S_IXUSR S_IEXEC
#endif

#if defined(__TURBOC__) && (__TURBOC__ <= 0x0296)
// Turbo/Borland C/C++ earlier than 3.10 doesn't have utime():
struct utimbuf {
  time_t actime;
  time_t modtime;
};

extern int utime (char* path, struct utimbuf* times);
#elif defined(__WATCOMC__)
// <Open> Watcom C/C++ has utime() in different header file:
  #include <sys/utime.h>
#else
  #include <utime.h>
#endif

#if defined(__TURBOC__) || defined(__WATCOMC__)
#include <io.h>
#endif

#include "stdio/frename.h"

int fcpy (const char* infile, const char* outfile, void* buf, size_t bufsize)
{
  FILE *fin=NULL, *fout=NULL;
  int res = 0, isallocated = 0;
  _stat_ statbuf;
  struct utimbuf times;

  if (!bufsize)
  {
    errno = ENOMEM;
    res = -1;
    goto lend;
  }

  if (buf == NULL)
  {
    buf = malloc (bufsize);
    if (buf == NULL)
    {
      errno = ENOMEM;
      res = -1;
      goto lend;
    }
    isallocated = 1;
  }

  fin = fopen (infile, "rb");
  fout = fopen (outfile, "wb");

  if ((fin == NULL) || (fout == NULL))
  {
    res = -1;
    goto lend;
  }

  for(;;)
  {
    size_t readcnt, writecnt;
    readcnt = fread (buf, 1, bufsize, fin);
    if (!readcnt)
      break;
    writecnt = fwrite (buf, 1, readcnt, fout);
    if (writecnt != readcnt)
    {
      res = -1;
      break;
    }
  }

  fclose (fin);
  fclose (fout);

  // get original date, time, attributes:
  stat ((char*)infile, &statbuf);

  // set exact date and time:
  times.actime = statbuf.st_atime;
  times.modtime = statbuf.st_mtime;
  utime ((char*)outfile, &times);

  // set exact writable (!read-only) attribute:
  if (!(statbuf.st_mode & S_IWUSR))
    chmod (outfile, 0);

lend:
  if (fin != NULL)
    fclose (fin);

  if (fout != NULL)
  {
    fclose (fout);
    if (res)
      remove(outfile);
  }

  if (isallocated)
    free (buf);
  return res;
}
