#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "FAT.h"
#include "image.h"
#include "testcase.h"

tBIOSDriveAndImageFName BIOSDriveAndImageFileName[] =
{
  {0x00, "FDDA.IMG"},
  {0x01, "FDDB.IMG"},
  {0x80, "HDD0.IMG"},
  {0x81, "HDD1.IMG"}
};

int main(int argc, char* argv[])
{
  uint index, startIndex = 0, stopIndex = UINT_MAX;
  uint ExecutedCnt = 0, PassedCnt = 0;
  int AllTestCases;
  int res = 1;

  if (argc != 2)
  {
lusage:
    printf ("\nUsage:\n"
            "  fattest <test case number>\n"
            "OR\n"
            "  fattest all\n");
    printf ("Test Cases:\n");
    for (index = 0; index < UINT_MAX; index++)
    {
      if ((aTestCases[index].pTextCaseTitle == NULL) ||
          (aTestCases[index].pTestCaseFxn == NULL))
        break;
      printf ("%5u: %s\n",
              aTestCases[index].TestCaseNo,
              aTestCases[index].pTextCaseTitle);
    }
    printf ("\n");
    return 1;
  }

  AllTestCases = (!strcmp(argv[1], "all")) || (!strcmp(argv[1], "ALL"));

  if (!AllTestCases)
  {
    uint no;

    if (sscanf (argv[1], "%u", &no) != 1)
    {
      printf ("Wrong test case number %s\n", argv[1]);
      goto lusage;
    }

    for (index = 0; index < UINT_MAX; index++)
    {
      if ((aTestCases[index].pTextCaseTitle == NULL) ||
          (aTestCases[index].pTestCaseFxn == NULL))
      {
        printf ("Wrong test case number %s\n", argv[1]);
        goto lusage;
      }
      if (aTestCases[index].TestCaseNo == no)
      {
        stopIndex = startIndex = index;
        break;
      }
    }
  }

  for (index = startIndex; index <= stopIndex; index++)
  {
    uint TestRes;

    if ((aTestCases[index].pTextCaseTitle == NULL) ||
        (aTestCases[index].pTestCaseFxn == NULL))
      break;

    ExecutedCnt++;

    printf ("Test case %u: %s\n",
            aTestCases[index].TestCaseNo,
            aTestCases[index].pTextCaseTitle);

    if (!stdio_InitFATFS(BIOSDriveAndImageFileName,
          sizeof(BIOSDriveAndImageFileName)/sizeof(BIOSDriveAndImageFileName[0]),
          0))
    {
      printf("Failed to \"mount\" one of FAT disk image files!\n");
    }
    else
    {
      TestRes = aTestCases[index].pTestCaseFxn();

      if (!stdio_DoneFATFS())
      {
        printf("Failed to unmount one of FAT disk image files!\n");
        TestRes = 1;
      }

      if (TestRes)
      {
        printf("Test case failed\n\n");
      }
      else
      {
        PassedCnt++;
        printf("Test case passed\n\n");
      }
    }
  }

  if (PassedCnt == ExecutedCnt)
    res = 0;

  printf ("\nTest cases:\nExecuted: %u, Passed: %u, Failed: %u\n\n",
          ExecutedCnt, PassedCnt, ExecutedCnt - PassedCnt);

  return res;
}
